
// How to generate and use band-limited steps with low-pass rolloff

#include "Wave_Writer.hpp"
#include <assert.h>
#include <stdio.h>
#include <math.h>

// Copyright (C) 2005 by Shay Green. MIT license.

// Number of phase offsets to sample band-limited step at
int const phase_count = 32;

// How many samples each final band-limited step contains
int const step_width = 16;

// Sine table

double const pi = 3.1415926535897932384626433832795029L;

// the +2 ensures that the last odd harmonic of the square wave can have a period of exactly 2
int const sine_size = 256 * phase_count + 2;
static float sine_table [sine_size]; // sin( 0 ) ... sin( pi * 2 )

static void init_sine_table()
{
	double const to_angle = pi * 2 / sine_size;
	for ( int i = 0; i < sine_size; i++ )
		sine_table [i] = sin( i * to_angle );
}

void add_sine( int angle_step, double amplitude, int count, float* out )
{
	assert( angle_step <= sine_size / 2 );
	int angle = 0;
	while ( count-- )
	{
		*out++ += sine_table [angle] * amplitude;
		angle += angle_step;
		if ( angle >= sine_size )
			angle -= sine_size;
	}
}

// Band-limited step

// band-limited step is symmetrical, so only one half needs to be generated
int const half_step_size = (step_width - 1) * phase_count / 2;
static float half_step [half_step_size];

void init_bl_step( double rolloff )
{
	init_sine_table();
	
	// generate very beginning of band-limited square
	double gain = 1.0 / 0.777; // adjust normal square wave's amplitude of 0.777 to 1.0
	for ( int h = 1; 1 * h <= sine_size / (phase_count * 2); h += 2 )
	{
		add_sine( h, gain / h, half_step_size, half_step );
		gain *= rolloff;
	}
}

// Step goes from 0.0 to 1.0, starting at -half_step_size and ending at half_step_size. Indicies
// outside this range return 0.0 and 1.0, respectively.
float bl_step( int index )
{
	double result = 0.5;
	if ( index < 0 )
	{
		result = -0.5;
		index = -index;
	} 
	
	if ( index < half_step_size )
		result *= half_step [index];
	
	return result + 0.5;
}

// Band-limited steps

// Would use short in a real program, for speed. Float here for clarity.
static float steps [phase_count] [step_width];

void init_steps( double rolloff )
{
	init_bl_step( rolloff );
	
	for ( int phase = 0; phase < phase_count; phase++ )
	{
		double error = 1.0;
		double prev = 0.0;
		for ( int i = 0; i < step_width; i++ )
		{
			double cur = bl_step( (i - step_width / 2 + 1) * phase_count + phase_count / 2 - phase );
			double delta = cur - prev;
			error -= delta;
			prev = cur;
			steps [phase] [i] = delta;
		}
		
		double half_error = error * 0.5;
		steps [phase] [step_width / 2 - 1] += half_error;
		steps [phase] [step_width / 2] += error - half_error;
		
		// print steps
		//for ( int j = 0; j < step_width; j++ )
		//  printf( "%f,", steps [phase] [j] );
		//printf( "\n" );
	}
}

int main()
{
	long const sample_rate = 44100;
	long i;
	
	// Initialize band-limited steps
	double rolloff = 0.998; // > 1.0 = emphasized highs, < 1.0 = reduced highs
	init_steps( rolloff );
	
	// Allocate and clear buffer
	long const buf_size = sample_rate * 4;
	float* buf = new float [buf_size];
	if ( !buf )
		return 1;
	for ( i = 0; i < buf_size; i++ )
		buf [i] = 0;
	
	// Add square wave of slowly lowering frequency
	double half_period = 32.04;
	double amp = 0.5;
	double last_amp = 0.0;
	int phase = 0;
	for ( double pos = 0; pos < buf_size - step_width; pos += half_period )
	{
		// Add band-limited step
		double delta = amp - last_amp;
		int whole = floor( pos );
		int phase = (pos - whole) * phase_count;
		for ( int i = 0; i < step_width; i++ )
			buf [whole + i] += steps [phase] [i] * delta;
		
		half_period *= 1.0001;
		last_amp = amp;
		amp = -amp;
	}
	
	// Integrate output buffer
	double sum = 0;
	for ( i = 0; i < buf_size; i++ )
		sum = (buf [i] += sum);
	
	// Write to wave file
	Wave_Writer wave( sample_rate );
	wave.write( buf, buf_size );
	
	return 0;
}

